import speech_recognition as sr
from .mic import Mic, MicConfig


class RecordingDevice:

    def __init__(self, mic_config: MicConfig) -> None:
        self.mic_config = mic_config
        self.mic = Mic(
            config=self.mic_config,
        )

        # We use SpeechRecognizer to record our audio because it has a nice
        # feature where it can detect when speech ends.
        self.recorder = sr.Recognizer()

        # TODO: Could add more config here.
        self.recorder.energy_threshold = self.mic_config.energy_threshold

        # Definitely do this, dynamic energy compensation lowers the energy threshold dramatically to a point where the SpeechRecognizer never stops recording.
        self.recorder.dynamic_energy_threshold = False

        with self.mic.source:
            self.recorder.adjust_for_ambient_noise(self.mic.source)



from dataclasses import dataclass
import logging


@dataclass
class LoggingConfig:
    """
    Logging configuration dataclass.

    level: The logging level to use.
    filepath: The filepath to write the log to.
    log_entry_format: The format of the log entry.
    date_format: The format of the date in the log entry.
    """
    level: int
    filepath: str
    log_entry_format: str
    date_format: str

    def __post_init__(self):
        level = getattr(logging, self.level.upper())
        logging.basicConfig(
            filename=self.filepath,
            format=self.log_entry_format,
            level=level,
            datefmt=self.date_format,
        )



from typing import Any, Callable, Dict, Optional
import numpy as np
import speech_recognition as sr
import whisper
import torch
from datetime import datetime, timedelta, timezone
from queue import Queue
from time import sleep
from rich import print
import logging

from .config import ListeningToolConfig
from .transcription import Segment, TranscriptionResult
from .recording_device import RecordingDevice


class ListeningTool:
    """
    A ListeningTool is a class that listens to audio input from a microphone and transcribes it into text.

    """

    def __init__(
        self,
        config: ListeningToolConfig,
        recording_device: RecordingDevice,
    ) -> None:

        self.config = config
        self.recording_device = recording_device
        self.recorder = sr.Recognizer()

        self.processing_rolling_avg_secs = 0.0

        print(f"Loading model: {self.config.transcribe_config.model}")
        self.audio_model = whisper.load_model(
            self.config.transcribe_config.model,
            self.config.transcribe_config.device,
            in_memory=self.config.in_memory,
        )

        # Thread safe Queue for passing data from the threaded recording callback.
        self.data_queue = Queue()

        self.transcription = [""]
        self.phrase_time = None

        # Create a background thread that will pass us raw audio bytes.
        # We could do this manually but SpeechRecognizer provides a nice helper.
        def record_callback(_, audio: sr.AudioData) -> None:
            """
            Threaded callback function to receive audio data when recordings finish.
            audio: An AudioData containing the recorded bytes.
            """
            # Grab the raw bytes and push it into the thread safe queue.
            data = audio.get_raw_data()
            self.data_queue.put(data)

        self.recording_device.recorder.listen_in_background(
            self.recording_device.mic.source,
            record_callback,
            phrase_time_limit=self.config.record_timeout,
        )

    def transcribe(self, audio_np: np.ndarray) -> TranscriptionResult:

        utc_dt = datetime.now(timezone.utc)  # UTC time
        local_datetime = utc_dt.astimezone()  # local time

        start_time = datetime.now()
        config = self.config.transcribe_config.to_dict()

        # TODO: Would be better if we broke out these config
        # into their own dataclass. Then, wouldn't need to delete them.
        del config["model"]
        del config["device"]
        del config["phrases_to_ignore"]

        result = self.audio_model.transcribe(
            audio_np,
            fp16=torch.cuda.is_available(),
            **config,
        )

        result["segments"] = self._deep_convert_np_float_to_float(result["segments"])

        segments = [Segment(**segment) for segment in result["segments"]]

        processing_secs = (datetime.now() - start_time).total_seconds()

        # TODO: Create rolling average window for processing time.
        self.processing_rolling_avg_secs = (
            (self.processing_rolling_avg_secs * 0.9 + processing_secs * 0.1)
            if self.processing_rolling_avg_secs
            else processing_secs
        )

        return TranscriptionResult(
            text=result["text"].strip(),
            segments=segments,
            language=result["language"],
            processing_secs=processing_secs,
            processing_rolling_avg_secs=self.processing_rolling_avg_secs,
            local_starttime=local_datetime,
        )

    def listen(self, callback: Optional[Callable[[str, Dict], None]] = None) -> None:
        while True:
            try:
                now = datetime.now(datetime.now().astimezone().tzinfo)

                # Pull raw recorded audio from the queue.
                if not self.data_queue.empty():

                    phrase_complete = False

                    # If enough time has passed between recordings, consider the phrase complete.
                    # Clear the current working audio buffer to start over with the new data.
                    if self._phrase_complete(self.phrase_time, now):
                        phrase_complete = True

                    # This is the last time we received new audio data from the queue.
                    self.phrase_time = now

                    # Combine audio data from queue
                    audio_data = b"".join(self.data_queue.queue)
                    self.data_queue.queue.clear()

                    # Convert in-ram buffer to something the model can use directly without needing a temp file.
                    # Convert data from 16 bit wide integers to floating point with a width of 32 bits.
                    # Clamp the audio stream frequency to a PCM wavelength compatible default of 32768hz max.
                    audio_np = (
                        np.frombuffer(audio_data, dtype=np.int16).astype(np.float32)
                        / 32768.0
                    )

                    # Read the transcription.
                    result = self.transcribe(audio_np)

                    # If we detected a pause between recordings, add a new item to our transcription.
                    # Otherwise edit the existing one.
                    if phrase_complete:
                        self.transcription.append(result.text)
                    else:
                        self.transcription[-1] = result.text

                    if (
                        callback
                        and result.text
                        not in self.config.transcribe_config.phrases_to_ignore
                    ):
                        if self.config.log:
                            logging.info(result.text)

                        callback(self.transcription, result)
                else:
                    # Infinite loops are bad for processors, must sleep.
                    sleep(0.25)
            except KeyboardInterrupt:
                break

    def _phrase_complete(self, phrase_time: datetime, now: datetime) -> bool:
        return phrase_time and now - phrase_time > timedelta(
            seconds=self.config.phrase_timeout
        )

    def _deep_convert_np_float_to_float(self, data: dict) -> dict:

        if isinstance(data, dict):
            for key, value in data.items():
                if isinstance(value, np.float64):
                    data[key] = float(value)
                if isinstance(value, list):
                    data[key] = self._deep_convert_np_float_to_float(value)
                elif isinstance(value, dict):
                    data[key] = self._deep_convert_np_float_to_float(value)
        if isinstance(data, list):
            for i, value in enumerate(data):
                if isinstance(value, np.float64):
                    data[i] = float(value)
                if isinstance(value, list):
                    data[i] = self._deep_convert_np_float_to_float(value)
                elif isinstance(value, dict):
                    data[i] = self._deep_convert_np_float_to_float(value)

        return data



from sys import platform
from dataclasses import dataclass
import speech_recognition
from rich import print


@dataclass
class MicConfig:
    mic_name: str
    sample_rate: int
    energy_threshold: int

    @classmethod
    def load(cls, data):
        return cls(**data)


@dataclass
class Mic:
    config: MicConfig
    source: speech_recognition.Microphone | None = None

    def __post_init__(self):

        if "linux" in platform:
            self._handle_linux()
        else:
            self.source = speech_recognition.Microphone(
                sample_rate=self.config.sample_rate
            )

    def _handle_linux(self):
        print("Available microphone devices are: ")
        for index, name in enumerate(
            speech_recognition.Microphone.list_microphone_names()
        ):
            print(f'Microphone with name "{name}" found')

        for index, name in enumerate(
            speech_recognition.Microphone.list_microphone_names()
        ):
            if self.config.mic_name in name:
                self.source = speech_recognition.Microphone(
                    sample_rate=self.config.sample_rate,
                    device_index=index,
                )
                print(f"Found target mic: '{self.config.mic_name}'")
                break

        if self.source is None:
            print(f"Target microphone not found: '{self.config.mic_name}'")
            print("Available devices are: ")
            for index, name in enumerate(
                speech_recognition.Microphone.list_microphone_names()
            ):
                print(f'Available microphones "{name}" found')
            quit("Exiting...")



import ast
from dataclasses import dataclass, asdict
from ast import literal_eval
from typing import List, Optional, Tuple, Union
from datetime import datetime

PHRASES_TO_IGNORE = [
    "",
    "urn.com urn.schemas-microsoft-com.h",
]


@dataclass
class Word:
    word: str
    start: float
    end: float
    probability: float

    @classmethod
    def load(cls, data):
        return cls(**data)

    def to_dict(self):
        return asdict(self)


@dataclass
class Segment:
    id: int
    seek: int
    start: float
    end: float
    text: str
    tokens: List[int]
    temperature: float
    avg_logprob: float
    compression_ratio: float
    no_speech_prob: float
    words: List[Word]

    def __post_init__(self):
        self.words = [Word(**word) for word in self.words]

    @classmethod
    def load(cls, data):
        return cls(**data)

    def to_dict(self):
        return asdict(self)


@dataclass
class TranscriptionResult:
    text: str
    segments: list[Segment]
    language: str
    processing_secs: int
    local_starttime: datetime
    processing_rolling_avg_secs: float = 0

    def __post_init__(self):
        for key, value in self.__dict__.items():
            if isinstance(value, datetime):
                self.__dict__[key] = value.isoformat()

    @classmethod
    def load(cls, data):
        return cls(**data)

    def to_dict(self):
        return asdict(self)


@dataclass
class TranscribeConfig:
    """
      transcribe_config:
    #  'tiny.en', 'tiny', 'base.en', 'base', 'small.en', 'small', 'medium.en', 'medium', 'large-v1', 'large-v2', 'large-v3', 'large', 'large-v3-turbo', 'turbo'
    model: medium

    # Whether to display the text being decoded to the console. If True, displays all the details, If False, displays minimal details. If None, does not display anything
    verbose: False

    # Temperature for sampling. It can be a tuple of temperatures,
    # which will be successively used upon failures according to
    # either compression_ratio_threshold or logprob_threshold.
    temperature: "(0.0, 0.2, 0.4, 0.6, 0.8, 1.0)" # "(0.0, 0.2, 0.4, 0.6, 0.8, 1.0)"

    # If the gzip compression ratio is above this value,
    # treat as failed
    compression_ratio_threshold: 2.4 # 2.4

    # If the average log probability over sampled tokens is below this value, treat as failed
    logprob_threshold: -1.0 # -1.0

    # If the no_speech probability is higher than this value AND
    # the average log probability over sampled tokens is below
    # logprob_threshold, consider the segment as silent
    no_speech_threshold: 0.6 # 0.6

    # if True, the previous output of the model is provided as a
    # prompt for the next window; disabling may make the text
    # inconsistent across windows, but the model becomes less
    # prone to getting stuck in a failure loop, such as repetition
    # looping or timestamps going out of sync.
    condition_on_previous_text: True # True

    # Extract word-level timestamps using the cross-attention
    # pattern and dynamic time warping, and include the timestamps
    # for each word in each segment.
    word_timestamps: False # False

    # If word_timestamps is True, merge these punctuation symbols
    # with the next word
    prepend_punctuations: >
      "\"'“¿([{-"

    # If word_timestamps is True, merge these punctuation symbols with the previous word
    append_punctuations: >
      "\"'.。,，!！?？:：”)]}、"

    # Optional text to provide as a prompt for the first window.
    # This can be used to provide, or "prompt-engineer" a context
    # for transcription, e.g. custom vocabularies or proper nouns
    # to make it more likely to predict those word correctly.
    initial_prompt: "" # ""

    # Keyword arguments to construct DecodingOptions instances
    decode_options: dict

    # Comma-separated list start,end,start,end,... timestamps
    # (in seconds) of clips to process. The last end timestamp
    # defaults to the end of the file.
    clip_timestamps: Union[str, List[float]]

    # When word_timestamps is True, skip silent periods longer
    # than this threshold (in seconds) when a possible
    # hallucination is detected
    hallucination_silence_threshold: ""
    """

    model: str
    device: str
    verbose: bool | None
    temperature: Union[float, Tuple[float, ...]]
    compression_ratio_threshold: float
    logprob_threshold: float
    no_speech_threshold: float
    condition_on_previous_text: bool
    word_timestamps: bool
    prepend_punctuations: str
    append_punctuations: str
    initial_prompt: Optional[str]
    clip_timestamps: Union[str, List[float]]
    hallucination_silence_threshold: Optional[float]
    phrases_to_ignore: list[str] = None

    @classmethod
    def load(cls, data):
        return cls(**data)

    def __post_init__(self):

        if self.phrases_to_ignore is None:
            self.phrases_to_ignore = PHRASES_TO_IGNORE

        if isinstance(self.verbose, str):
            self.verbose = literal_eval(self.verbose)
        if isinstance(self.temperature, str):
            self.temperature = literal_eval(self.temperature)
        if isinstance(self.clip_timestamps, str):
            if "," in self.clip_timestamps:
                values = self.clip_timestamps.split(",")
                self.clip_timestamps = (float(v) for v in values)
        if isinstance(self.hallucination_silence_threshold, str):
            self.hallucination_silence_threshold = literal_eval(
                self.hallucination_silence_threshold
            )
        if isinstance(self.temperature, str):
            self.temperature = ast.literal_eval(self.temperature)

        if isinstance(self.logprob_threshold, str):
            self.logprob_threshold = float(self.logprob_threshold)

        if isinstance(self.no_speech_threshold, str):
            self.no_speech_threshold = float(self.no_speech_threshold)

    def to_dict(self):
        return asdict(self)



from dataclasses import dataclass
from ast import literal_eval

import yaml
from .mic import MicConfig
from .logging_config import LoggingConfig
from .transcription import TranscribeConfig


@dataclass
class ListeningToolConfig:
    record_timeout: float
    phrase_timeout: float
    in_memory: bool
    log: bool
    transcribe_config: TranscribeConfig

    @classmethod
    def load(cls, data):
        return cls(**data)

    def __post_init__(self):
        self.transcribe_config = TranscribeConfig.load(self.transcribe_config)


@dataclass
class Config:
    listening_tool: ListeningToolConfig
    mic_config: MicConfig
    logging_config: LoggingConfig | None = None

    @classmethod
    def load(cls, path):
        with open(path, "r") as f:
            data = yaml.safe_load(f)
        return cls(**data)

    def __post_init__(self):
        self.listening_tool = ListeningToolConfig.load(self.listening_tool)
        self.mic_config = MicConfig.load(self.mic_config)
        if self.logging_config:
            self.logging_config = LoggingConfig(**self.logging_config)



from .logging_config import LoggingConfig
from .mic import MicConfig, Mic
from .recording_device import RecordingDevice
from .config import Config
from .transcription import TranscribeConfig, TranscriptionResult
from .listening_tool import ListeningTool
